@@ -18,36 +18,30 @@ calculatecentroid <- function(beta,returnlength = F){
   return(centroid)
 }
 
-
-innerprod_q2 <- function(q1, q2){
-    T1 = ncol(q1)
-    val = sum(q1*q2)/T1
-    return(val)
+innerprod_q2 <- function(q1, q2) {
+  T1 <- ncol(q1)
+  sum(q1 * q2) / T1
 }
 
-
-find_best_rotation <- function(q1, q2){
-    eps = .Machine$double.eps
-    n = nrow(q1)
-    T1 = ncol(q1)
-    A = q1%*%t(q2)
-    out = svd(A)
-    s = out$d
-    U = out$u
-    V = out$v
-    if (det(A)>0){
-        S = diag(1,n)
-    } else {
-        S = diag(1,n)
-        S[,n] = -S[,n]
+find_best_rotation <- function(q1, q2) {
+    eps <- .Machine$double.eps
+    L <- nrow(q1)
+    A <- q1 %*% t(q2)
+    out <- svd(A)
+    s <- out$d
+    U <- out$u
+    V <- out$v
+    if (det(A) > 0)
+      S <- diag(1, L)
+    else {
+      S <- diag(1, L)
+      S[, L] <- -S[, L]
     }
-    R = U%*%S%*%t(V)
-    q2new = R%*%q2
-
-    return(list(q2new=q2new, R=R))
+    R <- U %*% S %*% t(V)
+    q2new <- R %*% q2
+    list(q2new = q2new, R = R)
 }
 
-
 calculate_variance <- function(beta){
     n = nrow(beta)
     T1 = ncol(beta)
@@ -155,150 +149,186 @@ shift_f <- function(f, tau){
     return(fn)
 }
 
+find_rotation_seed_coord <- function(beta1, beta2,
+                                     mode = "O",
+                                     rotation = TRUE,
+                                     scale = TRUE,
+                                     lambda = 0.0) {
+  L <- nrow(beta1)
+  M <- ncol(beta1)
+  out <- curve_to_q(beta1, scale = scale)
+  q1 <- out$q
+  len1 <- out$len
 
-find_rotation_seed_coord <- function(beta1, beta2, mode="O", rotation=TRUE,
-                                     scale=TRUE){
-    n = nrow(beta1)
-    T1 = ncol(beta1)
-    q1 = curve_to_q(beta1, scale)$q
+  scl <- 4
+  minE <- Inf
+  if (mode == "C")
+    end_idx <- floor(M / scl)
+  else
+    end_idx <- 0
 
-    scl = 4
-    minE = 1000
-    if (mode=="C"){
-        end_idx = floor(T1/scl)
+  for (ctr in 0:end_idx) {
+    if (mode == "C") {
+      if (scl * ctr <= end_idx)
+        beta2n <- shift_f(beta2, scl * ctr)
+      else
+        break
+    } else
+      beta2n <- beta2
+
+    if (rotation) {
+      out <- find_best_rotation(beta1, beta2n)
+      beta2n <- out$q2new
+      Rout <- out$R
+    } else
+      Rout <- diag(nrow(beta2n))
+
+    out <- curve_to_q(beta2n, scale = scale)
+    q2n <- out$q
+    len2 <- out$len
+
+    if (norm(q1 - q2n, 'F') > 0.0001) {
+      if (scale) {
+        q1i <- q1
+        q2ni <- q2n
+      } else {
+        q1i <- q1 / sqrt(innerprod_q2(q1, q1))
+        q2ni <- q2n / sqrt(innerprod_q2(q2n, q2n))
+      }
+      dim(q1i) <- M * L
+      dim(q2ni) <- M * L
+      gam0 <- .Call('DPQ', PACKAGE = 'fdasrvf', q1i, q2ni,
+                    L, M, lambda, 1, 0, rep(0, M))
+      gamI <- invertGamma(gam0)
+      gam <- (gamI - gamI[1]) / (gamI[length(gamI)] - gamI[1])
+      beta2new <- group_action_by_gamma_coord(beta2n, gam)
+      out <- curve_to_q(beta2new, scale = scale)
+      q2new <- out$q
+      len2 <- out$len
+      if (mode == "C") {
+        q2new <- project_curve(q2new)
+        beta2new <- q_to_curve(q2new)
+      }
     } else {
-        end_idx = 0
+      q2new <- q2n
+      beta2new <- beta2n
+      gam <- seq(0, 1, length.out = M)
     }
 
-    for (ctr in 0:end_idx){
-        if (mode=="C"){
-            if ((scl*ctr) <= end_idx){
-              beta2n = shift_f(beta2, scl*ctr)
-            } else {
-              break
-            }
-        } else {
-            beta2n = beta2
-        }
+    Ec <- amplitude_distance(q1, q2new, scale = scale)
 
-        if (rotation){
-          out = find_best_rotation(beta1, beta2n)
-          beta2n = out$q2new
-          q2n = curve_to_q(beta2n, scale)$q
-          Rout = out$R
-        } else {
-          q2n = curve_to_q(beta2n, scale)$q
-          Rout = diag(nrow(beta2n))
-        }
-
-
-        if (norm(q1-q2n,'F') > 0.0001){
-            q1i = q1/sqrt(innerprod_q2(q1, q1))
-            q2ni = q2n/sqrt(innerprod_q2(q2n, q2n))
-            dim(q1i) = c(T1*n)
-            dim(q2ni) = c(T1*n)
-            gam0 = .Call('DPQ', PACKAGE = 'fdasrvf', q1i, q2ni, n, T1, 0, 1, 0, rep(0,T1))
-            gamI = invertGamma(gam0)
-            gam = (gamI-gamI[1])/(gamI[length(gamI)]-gamI[1])
-            beta2n = q_to_curve(q2n)
-            beta2new = group_action_by_gamma_coord(beta2n, gam)
-            q2new = curve_to_q(beta2new)$q
-            if (mode=="C"){
-                q2new = project_curve(q2new)
-            }
-        } else{
-            q2new = q2n
-            beta2new = beta2n
-            gam = seq(0,1,length.out=T1)
-        }
-        dist = innerprod_q2(q1,q2new)
-        if (dist < -1){
-            dist = -1
-        }
-        if (dist > 1){
-            dist = 1
-        }
-        Ec = acos(dist)
-        if (Ec < minE){
-            Rbest = Rout
-            beta2best = beta2new
-            q2best = q2new
-            gambest = gam
-            minE = Ec
-            tau = scl*ctr
-        }
+    if (Ec < minE) {
+      Rbest <- Rout
+      beta2best <- beta2new
+      q2best <- q2new
+      gambest <- gam
+      minE <- Ec
+      tau <- scl * ctr
     }
+  }
 
-    return(list(beta2best=beta2best,q2best=q2best,Rbest=Rbest,gambest=gambest,tau=tau))
+  ratio <- 1
+  if (scale)
+    ratio <- len1 / len2
+
+  list(
+    beta2best = beta2best * ratio,
+    q2best = q2best,
+    Rbest = Rbest,
+    gambest = gambest,
+    tau = tau,
+    scale = ratio
+  )
 }
 
+find_rotation_seed_unique <- function(q1, q2,
+                                      mode = "O",
+                                      alignment = TRUE,
+                                      rotation = TRUE,
+                                      scale = TRUE,
+                                      norm_ratio = 1.0,
+                                      lambda = 0.0) {
+  L <- nrow(q1)
+  M <- ncol(q1)
 
-find_rotation_seed_unqiue <- function(q1, q2, mode="O", rotation=TRUE, scale=TRUE,
-                                      lam=0.0){
-    n1 = nrow(q1)
-    T1 = ncol(q1)
-    scl = 4
-    minE = 1000
-    if (mode=="C"){
-        end_idx = floor(T1/scl)
+  # Variables for DPQ2 algorithm
+  grd <- seq(0, 1, length.out = M)
+  nbhd_dim <- 7L
+  Gvec <- rep(0, M)
+  Tvec <- rep(0, M)
+  size <- 0
+
+  scl <- 4
+  minE <- Inf
+
+  if (mode == "C")
+    end_idx <- floor(M / scl)
+  else
+    end_idx <- 0
+
+  for (ctr in 0:end_idx) {
+    if (mode == "C")
+      q2n <- shift_f(q2, scl * ctr)
+    else
+      q2n <- q2
+
+    if (rotation) {
+      out <- find_best_rotation(q1, q2n)
+      q2n <- out$q2new
+      Rbest <- out$R
+    } else
+      Rbest <- diag(nrow(q2n))
+
+    if (norm(q1 - q2n, 'F') > 0.0001 && alignment) {
+      if (scale) {
+        q1i <- q1
+        q2ni <- q2n
+      } else {
+        q1i <- q1 / sqrt(innerprod_q2(q1, q1))
+        q2ni <- q2n / sqrt(innerprod_q2(q2n, q2n))
+      }
+      dim(q1i) <- M * L
+      dim(q2ni) <- M * L
+
+      ret <- .Call(
+        "DPQ2", PACKAGE = "fdasrvf",
+        q1i, grd, q2ni, grd, L, M, M, grd, grd, M,
+        M, Gvec, Tvec, size, lambda, nbhd_dim
+      )
+      Gvec <- ret$G[1:ret$size]
+      Tvec <- ret$T[1:ret$size]
+      gamI <- stats::approx(Tvec, Gvec, xout = grd)$y
+
+      gam <- (gamI - gamI[1]) / (gamI[length(gamI)] - gamI[1])
+      q2new <- group_action_by_gamma(q2n, gam, scale = scale)
+
+      if (mode == "C")
+        q2new <- project_curve(q2new)
     } else {
-        end_idx = 0
+      q2new <- q2n
+      gam <- seq(0, 1, length.out = M)
     }
 
-    for (ctr in 0:end_idx){
-        if (mode=="C"){
-            q2n = shift_f(q2, scl*ctr)
-        } else {
-            q2n = q2
-        }
+    Ec <- amplitude_distance(q1, q2new, scale = scale, norm_ratio = norm_ratio)
 
-        if (rotation){
-          out = find_best_rotation(q1, q2n)
-          q2n = out$q2new
-          Rbest = out$R
-        } else {
-          Rbest = diag(nrow(q2n))
-        }
-
-
-        if (norm(q1-q2n,'F') > 0.0001){
-            q1i = q1/sqrt(innerprod_q2(q1, q1))
-            q2ni = q2n/sqrt(innerprod_q2(q2n, q2n))
-            dim(q1i) = c(T1*n1)
-            dim(q2ni) = c(T1*n1)
-            gam0 = .Call('DPQ', PACKAGE = 'fdasrvf', q1i, q2ni, n1, T1, lam, 1, 0, rep(0,T1))
-            gamI = invertGamma(gam0)
-            gam = (gamI-gamI[1])/(gamI[length(gamI)]-gamI[1])
-            beta2n = q_to_curve(q2n)
-            beta2new = group_action_by_gamma_coord(beta2n, gam)
-            q2new = curve_to_q(beta2new)$q
-            if (mode=="C"){
-                q2new = project_curve(q2new)
-            }
-        } else{
-            q2new = q2n
-            gam = seq(0,1,length.out=T1)
-        }
-        dist = innerprod_q2(q1,q2new)
-        if (dist < -1){
-            dist = -1
-        }
-        if (dist > 1){
-            dist = 1
-        }
-        Ec = acos(dist)
-        if (Ec < minE){
-            Rbest = Rbest
-            q2best = q2new
-            gambest = gam
-            minE = Ec
-        }
+    if (Ec < minE) {
+      Rbest <- Rbest
+      q2best <- q2new
+      gambest <- gam
+      minE <- Ec
+      tau <- scl * ctr
     }
+  }
 
-    return(list(q2best=q2best,Rbest=Rbest,gambest=gambest))
+  list(
+    q2best = q2best,
+    Rbest = Rbest,
+    gambest = gambest,
+    tau = tau,
+    d = minE
+  )
 }
 
-
 find_rotation_and_seed_q <- function(q1,q2){
     n = nrow(q1)
     T1 = ncol(q1)
@@ -319,35 +349,32 @@ find_rotation_and_seed_q <- function(q1,q2){
     return(list(q2new=q2new,O_hat=O_hat,tau=tau))
 }
 
+group_action_by_gamma <- function(q, gamma, scale = TRUE) {
+  L <- nrow(q)
+  M <- ncol(q)
+  grd <- seq(0, 1, length.out = M)
+  gammadot <- gradient(gamma, 1.0 / M)
+  qn <- matrix(nrow = L, ncol = M)
 
-group_action_by_gamma <- function(q, gamma){
-    n = nrow(q)
-    T1 = ncol(q)
-    gammadot = gradient(gamma, 1.0/T1)
-    qn = matrix(0, n, T1)
-    timet = seq(0, 1, length.out = T1)
+  for (l in 1:L)
+    qn[l, ] <- stats::spline(grd, q[l, ], xout = gamma)$y * sqrt(gammadot)
 
-    for (j in 1:n){
-        qn[j,] = stats::spline(timet, q[j,], xout=gamma)$y * sqrt(gammadot)
-    }
+  if (scale)
+    qn <- qn / sqrt(innerprod_q2(qn, qn))
 
-    qn = qn/sqrt(innerprod_q2(qn,qn))
-
-    return(qn)
+  qn
 }
 
+group_action_by_gamma_coord <- function(f, gamma) {
+  L <- nrow(f)
+  M <- ncol(f)
+  fn <- matrix(nrow = L, ncol = M)
+  grd <- seq(0, 1, length.out = M)
 
-group_action_by_gamma_coord <- function(f, gamma){
-    n = nrow(f)
-    T1 = ncol(f)
-    fn = matrix(0, n, T1)
-    timet = seq(0, 1, length.out = T1)
+  for (l in 1:L)
+    fn[l, ] <- stats::spline(grd, f[l, ], xout = gamma)$y
 
-    for (j in 1:n){
-        fn[j,] = stats::spline(timet, f[j,], xout=gamma)$y
-    }
-
-    return(fn)
+  fn
 }
 
 
@@ -609,22 +636,79 @@ rot_mat <- function(theta){
     return(O)
 }
 
+karcher_calc <- function(q1, mu, basis,
+                         rotated = TRUE,
+                         mode = "O",
+                         lambda = 0.0,
+                         ms = "mean") {
+    tmp <- dim(q1)
+    n <- tmp[1]
+    T1 <- tmp[2]
 
-karcher_calc <- function(beta, q, betamean, mu, rotated=T, mode="O"){
-    if (mode=="C"){
-        basis = find_basis_normal(mu)
-    }
-    # Compute shooting vector form mu to q_i
-    out = inverse_exp_coord(betamean, beta, mode, rotated)
+    out <- find_rotation_seed_unique(
+      q1 = mu,
+      q2 = q1,
+      mode = mode,
+      rotation = rotated,
+      scale = TRUE,
+      lambda = lambda
+    )
 
-    # Project to tangent space of manifold to obtain v_i
-    if (mode=="O"){
-        v = out$v
-    } else {
-        v = project_tangent(out$v, q, basis)
-    }
+    qn_t <- out$q2best / sqrt(innerprod_q2(out$q2best, out$q2best))
 
-    return(list(v=v,d=out$dist,gam=out$gam))
+    q1dotq2 <- innerprod_q2(mu, qn_t)
+
+    if (q1dotq2 >  1) q1dotq2 <-  1
+    if (q1dotq2 < -1) q1dotq2 <- -1
+
+    dist <- acos(q1dotq2)
+
+    u <- qn_t - q1dotq2 * q1
+    normu <- sqrt(innerprod_q2(u, u))
+    if (normu > 1e-4)
+      w <- u * acos(q1dotq2) / normu
+    else
+      w <- matrix(0, n, T1)
+
+    if (mode == "O")
+      v <- w
+    else
+      v <- project_tangent(w, q1, basis)
+
+    d_i <- 0
+    if (ms == "median") {
+      # run for median only, saves computation time if getting mean
+      d_i <- sqrt(innerprod_q2(v, v)) # calculate sqrt of norm of v_i
+      if (d_i > 0)
+        v_d <- v / d_i #normalize v_i
+      else
+        v_d <- v
+    } else
+      v_d <- matrix(0, n, T1)
+
+    list(v = v, v_d = v_d, dist = dist, d_i = d_i)
+}
+
+curve_align_sub <- function(beta1, q1, mu, mode, rotated, scale, lambda){
+  out = find_rotation_seed_unique(mu, q1, mode, rotated, TRUE, lambda)
+  gam = out$gambest
+  rotmat = out$Rbest
+
+  if (scale){
+    q1n = out$q2best
+    beta1n = q_to_curve(q1n)
+  } else {
+    beta1 = rotmat%*%beta1
+    beta1n = group_action_by_gamma_coord(beta1, gam)
+    q1n = curve_to_q(beta1n, scale)$q
+  }
+
+  T1 = ncol(beta1)
+  centroid1 = calculatecentroid(beta1n)
+  dim(centroid1) = c(length(centroid1), 1)
+  beta1n = beta1n - repmat(centroid1, 1, T1)
+
+  return(list(qn=q1n,betan=beta1n,rotmat=rotmat,gam=gam))
 }
 
 
@@ -641,3 +725,64 @@ elastic_shooting <- function(q1, v,mode="O"){
 
     return(q2n)
 }
+
+curve_to_srvf <- function(beta, scale = TRUE) {
+  centroid <- calculatecentroid(beta)
+  beta <- beta - centroid
+  out <- curve_to_q(beta, scale = scale)
+  list(q = out$q, qnorm = out$lenq, centroid = centroid)
+}
+
+match_f2_to_f1 <- function(srvf1, srvf2, beta2,
+                           mode = "O",
+                           alignment = TRUE,
+                           rotation = FALSE,
+                           scale = FALSE,
+                           include_length = FALSE,
+                           lambda = 0.0) {
+  norm_ratio <- 1
+  if (include_length)
+    norm_ratio <- srvf1$qnorm / srvf2$qnorm
+  out <- find_rotation_seed_unique(
+    srvf1$q, srvf2$q,
+    mode = mode,
+    alignment = alignment,
+    rotation = rotation,
+    scale = scale,
+    lambda = lambda,
+    norm_ratio = norm_ratio
+  )
+
+  # Compute amplitude distance
+  d <- out$d
+
+  # Compute phase distance
+  gam <- out$gambest
+  dx <- phase_distance(gam)
+
+  # Compute beta2n
+  qscale <- 1
+  if (scale)
+    qscale <- srvf1$qnorm / srvf2$qnorm
+  betascale <- qscale^2
+
+  if (mode == "C" && scale) {
+    beta2n <- q_to_curve(out$q2best, scale = qscale)
+  } else {
+    beta2n <- beta2 - srvf2$centroid
+    beta2n <- out$Rbest %*% beta2n
+    beta2n <- group_action_by_gamma_coord(beta2n, gam)
+    beta2n <- beta2n * betascale
+  }
+  beta2n <- beta2n + srvf1$centroid
+
+  list(
+    d = d,
+    dx = dx,
+    q2n = out$q2best,
+    beta2n = beta2n,
+    betascale = betascale,
+    Rbest = out$Rbest,
+    gambest = gam
+  )
+}
